using RaftModLoader;
﻿using UnityEngine;
using HMLLibrary;
using HarmonyLib;
using System.Linq;
using System.Collections.Generic;
using System.Runtime.CompilerServices;
using Steamworks;

public class PickupToStorage : Mod
{
    public static PickupToStorage Instance;
    Harmony harmony;

    // StorageManagerインスタンス
    private static StorageManager storageManager = null;

    // ユーザ設定
    public static bool useAllStorage = false;
    public static bool useEmptyStorage = true;
    public static List<string> targetItems = new List<string>();
    public static Dictionary<string, bool> DefaultItemSettings = new Dictionary<string, bool>()
    {
        { "Plank", true },
        { "Plastic", true },
        { "Thatch", true },
        { "Scrap", true },
        { "Stone", true },
        { "Rope", true },
        { "Raw_Potato", false },
        { "Raw_Beet", false },
    };


    public void Start()
    {
        CustomLogger.SetModName(GetModInfo().name);
        CustomLogger.EnableLogging = false;

        Instance = this;
        harmony = new Harmony("com.sora.PickupToStorage");
        harmony.PatchAll();

        Debug.Log($"[{GetModInfo().name}] Mod has been loaded!");
    }

    public void OnModUnload()
    {
        harmony.UnpatchAll(harmony.Id);
        Debug.Log($"[{GetModInfo().name}] Mod has been unloaded!");
    }

    // ブループリント入手デバッグ用
    //[HarmonyPatch(typeof(Pickup_Specific_Blueprint), "PickupSpecific")]
    //class Patch_Pickup_Specific_Blueprint
    //{
    //    private static float timePlayed = 0f;
    //    static void Prefix(Pickup_Specific_Blueprint __instance, ref PlayerInventory inventory)
    //    {
    //        timePlayed = GameManager.TimePlayed;

    //        var fieldInfo = AccessTools.Field(typeof(Pickup_Specific_Blueprint), "requiredTimePlayed");
    //        float requiredTimePlayed = (float)fieldInfo.GetValue(__instance);

    //        // プレイ時間を改ざんし、ブループリントを入手可能にする
    //        GameManager.TimePlayed = requiredTimePlayed * 60f + 1f;
    //    }

    //    static void Postfix()
    //    {
    //        // プレイ時間を元の正しい数値に戻す
    //        GameManager.TimePlayed = timePlayed;
    //    }
    //}

    [HarmonyPatch(typeof(Pickup), "AddItemToInventory")]
    class Patch_AddItemToInventory
    {
        // privateフィールド
        private static PlayerInventory playerInventory;
        private static InventoryPickup inventoryPickup;
        private static Network_Player playerNetwork;

        static bool Prefix(Pickup __instance, ref PickupItem item)
        {
            CustomLogger.Log("Prefix Start!");

            if (item == null)
                return true;

            // 拾ったアイテムをログ表示
            CustomLogger.Log($"PickupName: {item.PickupName}, name: {item.name}");
            if (item.itemInstance != null)
                CustomLogger.Log($"UniqueName: {item.itemInstance.UniqueName}, UniqueIndex: {item.itemInstance.UniqueIndex}");
            CustomLogger.Log($"pickupItemType: {item.pickupItemType}, pickupTerm: {item.pickupTerm}");
            if (item.yieldHandler != null)
                CustomLogger.Log($"Yield.Count: {item.yieldHandler.Yield.Count}, dropper: {item.dropper}");

            // 対象アイテムでなければ元のメソッドに任せる
            if (!IsFloatingTargetItem(item.name))
            {
                CustomLogger.Log("item.name is NOT my target. Prefix End!");
                return true;
            }

            // 条件想定外なら安全のため何もせず元のメソッドに任せる
            if (item.itemInstance != null && item.itemInstance.Valid)
            {
                CustomLogger.Log("Error: itemInstance is Valid. 想定外");
                return true;
            }
            if (item.yieldHandler == null)
            {
                CustomLogger.Log("Error: yeildHandler is null. 想定外");
                return true;
            }

            // privateフィールドを取得
            GetPrivateFields(__instance);

            if (item.yieldHandler.Yield.Count == 0 && item.dropper != null)
            {
                // dropperからアイテムを取り出し、一つずつ処理
                Item_Base[] randomItems = item.dropper.GetRandomItems();
                foreach (var randomItem in randomItems)
                {
                    if (randomItem != null)
                    {
                        CustomLogger.Log($"dropper dropped {randomItem.UniqueName}");

                        if (IsDroppedTargetItem(randomItem.UniqueName))
                        {
                            // 対象アイテムなので、ストレージに追加
                            AddSuitableInventories(__instance, randomItem.UniqueName, 1);
                        }
                        else
                        {
                            // 対象アイテムではないので、プレイヤーのインベントリに追加
                            int remainAmount = playerInventory.AddItem(randomItem.UniqueName, 1);
                            CustomLogger.Log($"Added to playerInventory. Remaining {remainAmount}.");
                        }
                    }
                }
            }
            else if (item.yieldHandler.Yield.Count > 0 && item.dropper == null)
            {
                // 対象アイテムの場合、Yield.Countは１だが念の為全部回して処理
                using (List<Cost>.Enumerator enumerator = item.yieldHandler.Yield.GetEnumerator())
                {
                    while (enumerator.MoveNext())
                    {
                        Cost cost = enumerator.Current;
                        CustomLogger.Log($"Yield item is {cost.amount} {cost.item.UniqueName}");
                        AddSuitableInventories(__instance, cost.item.UniqueName, cost.amount);
                    }
                }
            }
            else
            {
                CustomLogger.Log("Error: Yield.Count or dropper is unexpected. 想定外");
                return true;
            }

            // ブループリント等の特別なアイテムを取得
            if (item.specificPickups != null)
            {
                Pickup_Specific[] specificPickups = item.specificPickups;
                for (int j = 0; j < specificPickups.Length; j++)
                {
                    specificPickups[j].PickupSpecific(playerInventory);
                }
            }

            return false;
        }

        static bool IsFloatingTargetItem(string itemName)
        {
            // 樽と箱
            List<string> targetNames = new List<string>()
            {
                "Pickup_Floating_Barrel",
                "Pickup_Floating_Box",
            };

            // その他の対象アイテムを追加
            targetNames.AddRange(targetItems.Select(item => "Pickup_Floating_" + item).ToList());

            // 対象アイテムか判定
            bool hasPrefix = targetNames.Any(name => itemName.StartsWith(name));
            return hasPrefix;
        }

        static bool IsDroppedTargetItem(string itemUniqueName)
        {
            // 対象アイテムか判定
            bool hasMatch = targetItems.Any(name => itemUniqueName.Equals(name));
            return hasMatch;
        }

        static void GetPrivateFields(Pickup __instance)
        {
            // playerInventoryフィールドを取得
            var fieldInfo = AccessTools.Field(typeof(Pickup), "playerInventory");
            playerInventory = (PlayerInventory)fieldInfo.GetValue(__instance);

            // inventoryPickupフィールドを取得
            var inventoryPickupFieldInfo = AccessTools.Field(typeof(PlayerInventory), "inventoryPickup");
            inventoryPickup = (InventoryPickup)inventoryPickupFieldInfo.GetValue(playerInventory);

            // playerNetworkフィールドを取得
            if (storageManager == null)
                storageManager = FindObjectOfType<StorageManager>();
            if (storageManager == null)
            {
                CustomLogger.LogWarning("storageManager is null.");
                return;
            }
            var playerNetworkField = AccessTools.Field(typeof(StorageManager), "playerNetwork");
            playerNetwork = (Network_Player)playerNetworkField.GetValue(storageManager);
        }

        static int AddSuitableInventories(Pickup __instance , string itemUniqueName, int itemAmount)
        {
            // 初期状態の拾った個数を覚えておく
            int pickupAmount = itemAmount;

            // ストレージに追加していく
            int remainAmount = AddToStorages(itemUniqueName, itemAmount);

            if (remainAmount == 0)
            {
                // 残りがなければ、最初に拾った数をUI表示
                inventoryPickup.ShowItem(itemUniqueName, pickupAmount);
            }
            else
            {
                CustomLogger.Log($"{remainAmount} {itemUniqueName} ovrflowed from storages.");

                // 残りのアイテムがあれば、プレイヤーのインベントリに追加
                remainAmount = playerInventory.AddItem(itemUniqueName, remainAmount);
                CustomLogger.Log($"Added to playerInventory. Remaining {remainAmount}.");
            }
            
            return remainAmount;
        }

        static int AddToStorages(string itemUniqueName, int itemAmount)
        {
            // 全ストレージをチェックし、対象アイテムがあれば追加していく
            foreach (var storage in StorageManager.allStorages)
            {
                CustomLogger.Log($"storage.name: {storage.name}, ObjectIndex: {storage.ObjectIndex}");

                // 対象ストレージでなければスキップ
                if (!IsTargetStorage(storage.name))
                {
                    CustomLogger.Log("storage.name is NOT my target.");
                    continue;
                }

                Inventory inventory = storage.GetInventoryReference();

                if (inventory == null)
                {
                    CustomLogger.Log($"Error: inventory is null. 想定外");
                    continue;
                }

                // このストレージに対象アイテムがあれば追加
                if (ContainsItem(inventory, itemUniqueName))
                {
                    itemAmount = inventory.AddItem(itemUniqueName, itemAmount);
                    SyncStorageState(storage);
                    CustomLogger.Log($"Add item to storage. remaining amount: {itemAmount}");
                    if (itemAmount == 0)
                        return itemAmount;
                }
            }

            // 空のストレージを使用する設定でなければここで終了
            if (!useEmptyStorage)
                return itemAmount;

            // 全ストレージをチェックし、空のストレージがあれば追加
            foreach (var storage in StorageManager.allStorages)
            {
                CustomLogger.Log($"storage.name: {storage.name}, ObjectIndex: {storage.ObjectIndex}");

                // 対象ストレージでなければスキップ
                if (!IsTargetStorage(storage.name))
                {
                    CustomLogger.Log("storage.name is NOT my target.");
                    continue;
                }

                Inventory inventory = storage.GetInventoryReference();

                if (inventory == null)
                {
                    CustomLogger.Log($"Error: inventory is null. 想定外");
                    continue;
                }

                // 空のストレージなら追加
                if (IsEmptyStorage(inventory))
                {
                    itemAmount = inventory.AddItem(itemUniqueName, itemAmount);
                    SyncStorageState(storage);
                    CustomLogger.Log($"Add item to empty storage. remaining amount: {itemAmount}");
                    break;
                }
            }

            // 残りの数を返す
            return itemAmount;
        }

        static void SyncStorageState(Storage_Small currentStorage)
        {
            if (currentStorage == null)
            {
                CustomLogger.Log("Error: currentStorage is null. 想定外");
                return;
            }

            if (playerNetwork == null)
            {
                CustomLogger.LogWarning("playerNetwork is null.");
                return;
            }

            // ストレージの中身を他の人に更新させるために、Closeの通知を送る
            Message_Storage_Close message = new Message_Storage_Close(Messages.StorageManager_Close, storageManager, currentStorage);
            if (Raft_Network.IsHost)
            {
                // ホストの場合、RPCでStorageManager_Closeを通知
                playerNetwork.Network.RPC(message, Target.Other, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
                return;
            }
            // クライアントの場合、P2P通信でStorageManager_Closeを通知
            playerNetwork.SendP2P(message, EP2PSend.k_EP2PSendReliable, NetworkChannel.Channel_Game);
            return;
        }

        static bool IsTargetStorage(string storageName)
        {
            // 対象ストレージか判定
            if (useAllStorage)
            {
                // ゴミ箱のみ除外
                return !storageName.StartsWith("Placeable_Trashcan");
            }
            else
            {
                // 以下3種類のストレージのみ可
                List<string> targetStorages = new List<string>()
                {
                    "Placeable_Storage_Medium",
                    "Placeable_Storage_Small",
                    "Placeable_Storage_Large",
                };

                bool hasPrefix = targetStorages.Any(name => storageName.StartsWith(name));
                return hasPrefix;
            }
        }

        static bool ContainsItem(Inventory inventory, string uniqueItemName)
        {
            // すべてのスロットを線形検索して、対象アイテムの存在を確認
            foreach (Slot slot in inventory.allSlots)
            {
                if (!slot.IsEmpty && slot.itemInstance.UniqueName == uniqueItemName)
                {
                    // 対象アイテムが見つかった
                    return true;
                }
            }
            // 全スロットを検索しても対象アイテムが見つからなかった
            return false;
        }

        static bool IsEmptyStorage(Inventory inventory)
        {
            // すべてのスロットを線形検索して、空か判定
            foreach (var slot in inventory.allSlots)
            {
                if (!slot.IsEmpty)
                    return false;
            }
            return true;
        }

        static void DebugPrintAllStorages()
        {
            // 全ストレージ表示
            CustomLogger.Log($"allStorages.Count: {StorageManager.allStorages.Count}");
            foreach (var storage in StorageManager.allStorages)
            {
                //Inventory inventory = storage.GetComponent<Inventory>();
                //CustomLogger.Log($"inventory: {inventory}");
                //Inventory inventoryRef = storage.GetInventoryReference();
                //CustomLogger.Log($"inventoryRef: {storage}");

                CustomLogger.Log($"storage.ObjectIndex: {storage.ObjectIndex}");
                //CustomLogger.Log($"storage.dpsType: {storage.dpsType}");
                //CustomLogger.Log($"storage.GetType: {storage.GetType()}");
                //CustomLogger.Log($"storage.inventoryPrefab: {storage.inventoryPrefab}");
                CustomLogger.Log($"storage.name: {storage.name}");
                CustomLogger.Log($"storage.gameObject.name: {storage.gameObject.name}");
            }
        }
    }

    // ExtraSettingsAPI

    public void ExtraSettingsAPI_Load()
    {
        LoadSettingValues();
        CustomLogger.Log("Loaded existing setting values.");
    }

    public void ExtraSettingsAPI_SettingsClose()
    {
        LoadSettingValues();
        CustomLogger.Log("Settings menu closed. updated settings values.");
    }

    public void ExtraSettingsAPI_ButtonPress(string SettingName)
    {
        if (SettingName == "Reset")
        {
            ResetSettingValues();
            CustomLogger.Log("Reset all settings.");
        }
    }

    private void LoadSettingValues()
    {
        useAllStorage = ExtraSettingsAPI_GetCheckboxState("UseAllStorage");
        useEmptyStorage = ExtraSettingsAPI_GetCheckboxState("UseEmptyStorage");
        
        // 対象アイテムのリストをクリアしてから追加
        targetItems.Clear();
        foreach (var item in DefaultItemSettings)
        {
            if (ExtraSettingsAPI_GetCheckboxState(item.Key))
                targetItems.Add(item.Key);
        }
    }

    private static void ResetSettingValues()
    {
        useAllStorage = false;
        useEmptyStorage = true;
        ExtraSettingsAPI_SetCheckboxState("UseAllStorage", false);
        ExtraSettingsAPI_SetCheckboxState("UseEmptyStorage", true);

        // 対象アイテムのリストをクリアしてから追加し、設定値も更新
        targetItems.Clear();
        foreach (var item in DefaultItemSettings)
        {
            if (item.Value)
                targetItems.Add(item.Key);
            ExtraSettingsAPI_SetCheckboxState(item.Key, item.Value);
        }
    }

    //| Because of some issues with how things are
    //| handled, it's highly recommended, but not
    //| required, to include the MethodImpl attribute
    //| on these methods

    // Use to get the current state of a Checkbox type setting
    [MethodImpl(MethodImplOptions.NoInlining)]
    public static bool ExtraSettingsAPI_GetCheckboxState(string SettingName) => false;

    // Use to set the current state of a Checkbox type setting
    [MethodImpl(MethodImplOptions.NoInlining)]
    public static void ExtraSettingsAPI_SetCheckboxState(string SettingName, bool value) { }
}

public static class CustomLogger
{
    public static bool EnableLogging = true;
    private static string modName = "";

    public static void SetModName(string name)
    {
        modName = "[" + name + "] ";
    }

    public static void Log(object message)
    {
        if (EnableLogging)
        {
            Debug.Log(modName + message);
        }
    }

    public static void LogWarning(object message)
    {
        if (EnableLogging)
        {
            Debug.LogWarning(modName + message);
        }
    }

    public static void LogError(object message)
    {
        if (EnableLogging)
        {
            Debug.LogError(modName + message);
        }
    }
}
